/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.addons.netviewer;

import java.awt.Color;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import monalisa.data.PropertyList;

/**
 * A Edge in the Net Viewer
 * @author Jens Einloft
 */

public class NetViewerEdge implements Serializable {
    private static final long serialVersionUID = 8357628722035312827L;

    private final String name;
    private int weight;
    private NetViewerNode source, aim;
    private NetViewerEdge masterEdge;
    private List<NetViewerEdge> bendEdges = new ArrayList<NetViewerEdge>();
    private Boolean visible;
    private Color color;

    private PropertyList properties;

    /**
     * Generate a new NetViewerEdge
     * @param name
     * @param weight
     * @param source NetViewerNode
     * @param aim NetViewerNode
     */
    NetViewerEdge(String name, int weight, NetViewerNode source, NetViewerNode aim) {
        this.name = name;
        this.weight = weight;
        this.source = source;
        this.aim = aim;
        this.visible = true;
        this.masterEdge = this;
        source.addOutEdge(this);
        aim.addInEdge(this);
        this.color = Color.BLACK;

        this.properties = new PropertyList();
    }

     /**
     * Generate a new NetViewerEdge
     * @param name
     * @param weight
     * @param source NetViewerNode
     * @param aim NetViewerNode
     * @param color
     */
    NetViewerEdge(String name, int weight, NetViewerNode source, NetViewerNode aim, Color color) {
        this.name = name;
        this.weight = weight;
        this.source = source;
        this.aim = aim;
        this.visible = true;
        this.masterEdge = this;
        source.addOutEdge(this);
        aim.addInEdge(this);
        this.color = color;

        this.properties = new PropertyList();
    }

    /**
     * Generate a new Bend Edge
     * @param name
     * @param weight
     * @param source
     * @param aim
     * @param masterEdge
     */
    NetViewerEdge(String name, int weight, NetViewerNode source, NetViewerNode aim, NetViewerEdge masterEdge) {
        this.name = name;
        this.weight = weight;
        this.source = source;
        this.aim = aim;
        this.visible = true;
        source.addOutEdge(this);
        aim.addInEdge(this);
        this.masterEdge = masterEdge;
        masterEdge.addBendEdge(this);
        this.color = Color.BLACK;

        this.properties = new PropertyList();
    }

    /**
     * Returns the name of the edge
     * @return
     */
    public String getName() {
        return this.name;
    }

    /**
     * returns the weight of the edge
     * @return
     */
    public int getWeight() {
        return this.weight;
    }

    public void setWeightForAllEdges(int weight) {
        this.weight = weight;
        for(NetViewerEdge e : this.masterEdge.bendEdges) {
            e.setWeight(weight);
        }
    }

    private void setWeight(int weight) {
        this.weight = weight;
    }

    /**
     * Returns the source of the edge
     * @return
     */
    public NetViewerNode getSource() {
        return this.source;
    }

    public void changeSource(NetViewerNode source) {
        this.source = source;
    }

    /**
     * Returns the aim of the edge
     * @return
     */
    public NetViewerNode getAim() {
        return this.aim;
    }

    public void changeAim(NetViewerNode aim) {
        this.aim = aim;
    }

    /**
     * Hidde or show the edge
     * @param value
     */
    public void setVisible(Boolean value) {
        this.visible = value;
    }

    /**
     * Returns the visibile of the edge
     * @return
     */
    public Boolean getVisible() {
        return this.visible;
    }

    /**
     * Returns the master edge of an edge
     * @return
     */
    public NetViewerEdge getMasterEdge() {
        return this.masterEdge;
    }

    /**
     * Add a bend edge to the list of bend edges
     * @param bendEdge
     */
    public void addBendEdge(NetViewerEdge bendEdge) {
        this.bendEdges.add(bendEdge);
    }

    /**
     * remove a bend egde from the list of bend edges
     * @param bendEdge
     */
    public void removeBendEdge(NetViewerEdge bendEdge) {
        this.bendEdges.remove(bendEdge);
    }

    /**
     * Returns the list of bend edges
     * @return
     */
    public List<NetViewerEdge> getBendEdges() {
        return this.bendEdges;
    }

    /**
     * Clear the list of bend edges
     */
    public void clearBendEdges() {
        this.bendEdges.clear();
    }

    /**
     * Set the color of an edge
     * @param color
     */
    private void setColor(Color color) {
        this.color = color;
    }

    /**
     * Set the color of all bend edges
     * @param color
     */
   public void setColorForAllEdges(Color color) {
        this.color = color;
        for(NetViewerEdge e : this.masterEdge.bendEdges) {
            e.setColor(color);
        }
    }

    /**
     * returns the color of an edge
     * @return
     */
    public Color getColor() {
        return this.color;
    }

        /**
     * Add a property to the Petri net entity.
     * @param <T> The type of the property.
     * @param key The key of the property.
     * @param value The property.
     */
    public <T> void putProperty(String key, T value) {
        properties.put(key, value);
    }

    /**
     * Retrieve a strongly typed property, based on its key.
     * @param <T> The type of the property to retrieve.
     * @param key The key of the property.
     * @return The property, cast to type <code>T</code>.
     */
    public <T> T getProperty(String key) {
        return properties.<T>get(key);
    }

     /**
     * Test whether the entity has a given property.
     * @param key The key to look for.
     * @return <code>true</code>, if the key is present, otherwise <code>false</code>.
     */
    public boolean hasProperty(String key) {
        return properties.has(key);
    }

    @Override
    public String toString() {
        return this.name;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this)
            return true;
        else if (obj.getClass().equals(this.getClass()))
                return this.getName().equalsIgnoreCase(((NetViewerEdge)obj).getName());
        else
            return false;
    }

    @Override
    public int hashCode() {
        return this.name.hashCode();
    }
}
