/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.input;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import monalisa.data.pn.Arc;
import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.util.FileUtils;

/**
 * Input handler for the Pipe4 PNML format.
 * @author Jens Einloft
 * @see <a href="http://pipe2.sourceforge.net/">Platform Independent Petri net Editor 4</a>
 **/
public class Pipe4InputHandler implements InputHandler {
 private final Map<String, Place> places = new HashMap<String, Place>();
    private final Map<String, Transition> transitions = new HashMap<String, Transition>();

    @Override
    public boolean isKnownFile(File file) throws IOException {
        if (!"xml".equalsIgnoreCase(FileUtils.getExtension(file)))
            return false;

        SAXBuilder builder = new SAXBuilder();
        Document doc;
        try {
            doc = builder.build(file);
        } catch (JDOMException e) {
            return false;
        }
        Element root = doc.getRootElement();

        // Pipe 4 uses PNML but doesn't a proper net type. We assume P/T
        // networks.

        if (!root.getName().equals("pnml"))
            return false;
        Element netNode = root.getChild("net");
        if (netNode == null)
            return false;
        // Pipe4?
        if (!netNode.getChildren("token").isEmpty())
            return true;
        else
            return false;
    }

    @SuppressWarnings("unchecked")
    @Override
    public PetriNet load(InputStream in) throws IOException {
        PetriNet ret = new PetriNet();

        SAXBuilder builder = new SAXBuilder();
        Document doc = null;
        try {
            doc = builder.build(in);
        } catch (JDOMException ex) {
            Logger.getLogger(Pipe3InputHandler.class.getName()).log(Level.SEVERE, null, ex);
        }

        // - pnml
        // -- net
        // --- (page) optional
        // ---- place
        // ---- transition
        // ---- arc

        List<Element> pnmlChildren = doc.getRootElement().getChildren();
        List<Element> workspace = null;

        // The <page> tag is optional
        if (pnmlChildren.get(0).getChildren().size() == 1) {
            workspace = pnmlChildren.get(0).getChildren();
        }  else {
            workspace = pnmlChildren;
        }

        // Net structure section
        int countPlaces = 0;
        int countTransitions = 0;
        String id,name;
        Long tokens;
        Double posX,posY;
        Object from, to;
        for(Element e : workspace){
            List<Element> elements = e.getChildren();
            for(Element pe : elements){
                // places
                if (pe.getName().equals("place")) {
                    id = pe.getAttributeValue("id");
                    name = pe.getChild("name").getChild("value").getValue();
                    tokens = Long.parseLong(pe.getChild("initialMarking").getChild("value").getValue().split(",")[1]);
                    posX = new Double(pe.getChild("graphics").getChild("position").getAttributeValue("x"));
                    posY = new Double(pe.getChild("graphics").getChild("position").getAttributeValue("y"));
                    if (name == null)
                        name = id;
                    Place place = findPlace(countPlaces, id, ret);
                    place.putProperty("name", name);
                    place.putProperty("posX", posX);
                    place.putProperty("posY", posY);
                    ret.setTokens(place, tokens);
                    countPlaces++;
                }
                // transitions
                else if (pe.getName().equals("transition")) {
                    id = pe.getAttributeValue("id");
                    name = pe.getChild("name").getChild("value").getValue();
                    posX = new Double(pe.getChild("graphics").getChild("position").getAttributeValue("x"));
                    posY = new Double(pe.getChild("graphics").getChild("position").getAttributeValue("y"));
                    if (name == null)
                        name = id;
                    Transition transition = findTransition(countTransitions, id, ret);
                    transition.putProperty("name", name);
                    transition.putProperty("posX", posX);
                    transition.putProperty("posY", posY);
                    countTransitions++;
                }
                // arcs
                else if(pe.getName().equals("arc")) {
                    int weight = Integer.parseInt(pe.getChild("inscription").getChild("value").getValue().split(",")[1]);
                    String source = pe.getAttributeValue("source");
                    String target = pe.getAttributeValue("target");
                    if (places.containsKey(source)) {
                        from = places.get(source);
                        to = transitions.get(target);
                        ret.addArc((Place)from, (Transition)to, new Arc(from, to, weight));
                    } else if (transitions.containsKey(source))  {
                        from = transitions.get(source);
                        to = places.get(target);
                        ret.addArc((Transition)from, (Place)to, new Arc(from, to, weight));
                    }
                }
            }
        }

        return ret;
    }

    private Place findPlace(int placeId, String stringId, PetriNet petriNet) {
        Place place = places.get(stringId);

        if (place == null) {
            place = new Place(placeId);
            places.put(stringId, place);
            petriNet.addPlace(place);
        }
        return place;
    }

    private Transition findTransition(int transitionId, String stringId, PetriNet petriNet) {
        Transition transition = transitions.get(stringId);

        if (transition == null) {
            transition = new Transition(transitionId);
            transitions.put(stringId, transition);
            petriNet.addTransition(transition);
        }
        return transition;
    }
}
