/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.pn;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.*;
import monalisa.data.Pair;

/**
 * <p>A Petri net model.</p>
 * <p><strong>Note</strong>:
 * For performance reasons, minimal sanity checks are performed. As a
 * consequence, many forbidden operations yield undefined behavior rather than
 * failing gracefully with an appropriate exception.
 * As an example, trying to retrieve the tokens of a non-existing place will
 * yield an undefined result, as will calling <code>addArc</code> etc.</p>
 * <p><strong>Attention!</strong> The API of this class is subject to change!
 * Future versions of this class will be immutable to change, a builder class
 * will provide functionality to build and change Petri nets. Do not rely on
 * this preliminary API in external projects.</p>
 * @author Konrad Rudolph & Jens Einloft
 */
public final class PetriNet extends AbstractPetriNetEntity {
    private static final long serialVersionUID = -8266249373834650762L;

    private Map<Integer, Place> places = new HashMap<Integer, Place>();
    private Map<Integer, Transition> transitions = new HashMap<Integer, Transition>();
    private Map<Place, Long> marking = new HashMap<Place, Long>();
    private Map<Place, List<Transition>> inputs = new HashMap<Place, List<Transition>>();
    private Map<Transition, List<Place>> outputs = new HashMap<Transition, List<Place>>();
    private Map<Pair<Place, Transition>, Arc> inarcs = new HashMap<Pair<Place,Transition>, Arc>();
    private Map<Pair<Transition, Place>, Arc> outarcs = new HashMap<Pair<Transition,Place>, Arc>();

    // Inverse lookup tables.
    private Map<Place, List<Transition>> preTransitions = new HashMap<Place, List<Transition>>();
    private Map<Transition, List<Place>> prePlaces = new HashMap<Transition, List<Place>>();

    /**
     * Creates a new, empty Petri net.
     */
    public PetriNet() { }
    
    /**
     * Creates a copy of the specified Petri net.
     * Performs a deep copy, i.e. creates copies of all entities in the Petri
     * net. Properties, on the other hand, are shared between the entities of
     * this Petri net and the other Petri net since there is no generic way of
     * copying properties efficiently (i.e. without serialization).
     * <strong>Note:</strong> This behavior is subject to change since it may
     * prove impractical.
     * @param other The Petri net to copy.
     */
    public PetriNet(PetriNet other) {
        super(other);
        
        for (Place oldPlace : other.places()) {
            Place newPlace = new Place(oldPlace);
            addPlace(newPlace);
            setTokens(newPlace, other.getTokens(oldPlace));
        }
        
        for (Transition oldTransition : other.transitions()) {
            Transition newTransition = new Transition(oldTransition);
            addTransition(newTransition);
            
            for (Place oldPlace : oldTransition.inputs()) {
                Place newPlace = findPlace(oldPlace.id());
                Arc newArc = new Arc(other.getArc(oldPlace, oldTransition));
                addArc(newPlace, newTransition, newArc);
            }
            
            for (Place oldPlace : oldTransition.outputs()) {
                Place newPlace = findPlace(oldPlace.id());
                Arc newArc = new Arc(other.getArc(oldTransition, oldPlace));
                addArc(newTransition, newPlace, newArc);
            }
        }
    }
    
    /**
     * Add a new place to the Petri net.
     * @param place The place.
     */
    public void addPlace(Place place) {
        places.put(place.id(), place);
        place.setContainer(this);
        inputs.put(place, new ArrayList<Transition>());
        preTransitions.put(place, new ArrayList<Transition>());
        marking.put(place, 0L);
    }
    
    /**
     * Add a list of new places to the Petri net.
     * @param places The places.
     */
    public void addPlaces(Iterable<Place> places) {
        for (Place place : places)
            addPlace(place);
    }
    
    /**
     * Add a list of new places to the Petri net.
     * @param places The places.
     */
    public void addPlaces(Place... places) {
        for (Place place : places)
            addPlace(place);
    }
    
    /**
     * Add a new transition to the Petri net.
     * @param transition The transition.
     */
    public void addTransition(Transition transition) {
        transitions.put(transition.id(), transition);
        transition.setContainer(this);
        outputs.put(transition, new ArrayList<Place>());
        prePlaces.put(transition, new ArrayList<Place>());
    }
    
    /**
     * Add a list of new transitions to the Petri net.
     * @param transitions The transitions.
     */
    public void addTransitions(Iterable<Transition> transitions) {
        for (Transition transition : transitions)
            addTransition(transition);
    }
    
    /**
     * Add a list of new transitions to the Petri net.
     * @param transitions The transitions.
     */
    public void addTransitions(Transition... transitions) {
        for (Transition transition : transitions)
            addTransition(transition);
    }
    
    /**
     * Add a new default arc from {@code from} to {@code to} the Petri net.
     * Corresponds to the following call:
     * <pre>addArc(from, to, new Arc());</pre>
     * @param from The starting place.
     * @param to The target transition.
     */
    public void addArc(Place from, Transition to) {
        addArc(from, to, new Arc(from, to));
    }
    
    public void addArc(Place from, Transition to, int weight) {
        addArc(from, to, new Arc(from, to, weight));
    }
    
    /**
     * Add a new custom arc from {code from} to {@code to} the Petri net.
     * @param from The starting place.
     * @param to The target transition.
     * @param arc The custom arc to add.
     */
    public void addArc(Place from, Transition to, Arc arc) {
        inputs.get(from).add(to);
        inarcs.put(Pair.of(from, to), arc);
        prePlaces.get(to).add(from);
    }
    
    /**
     * Add a new default arc from {code from} to {@code to} the Petri net.
     * Corresponds to the following call:
     * <pre>addArc(from, to, new Arc());</pre>
     * @param from The starting transition.
     * @param to The target place.
     */
    public void addArc(Transition from, Place to) {
        addArc(from, to, new Arc(from, to));
    }
    
    public void addArc(Transition from, Place to, int weight) {
        addArc(from, to, new Arc(from, to, weight));
    }
    
    /**
     * Add a new custom arc from {@code from} to {@code to} the Petri net.
     * @param from The starting transition.
     * @param to The target place.
     * @param arc The custom arc to add.
     */
    public void addArc(Transition from, Place to, Arc arc) {
        outputs.get(from).add(to);
        outarcs.put(Pair.of(from, to), arc);
        preTransitions.get(to).add(from);
    }

    public void removeArc(Place source, Transition aim) {
            if(outputs.containsKey(aim))
                outputs.get(aim).remove(source);
            if(inputs.containsKey(source))
                inputs.get(source).remove(aim);
            inarcs.remove(new Pair(source, aim));
            outarcs.remove(new Pair(aim, source));
            if(preTransitions.containsKey(source))
                preTransitions.get(source).remove(aim);
            if(prePlaces.containsKey(aim))
                prePlaces.get(aim).remove(source);
    }

    public void removeArc(Transition source, Place aim) {
            if(outputs.containsKey(source))
                outputs.get(source).remove(aim);
            if(inputs.containsKey(aim))
                inputs.get(aim).remove(source);
            inarcs.remove(new Pair(aim, source));
            outarcs.remove(new Pair(source, aim));
            if(preTransitions.containsKey(aim))
                preTransitions.get(aim).remove(source);
            if(prePlaces.containsKey(source))
                prePlaces.get(source).remove(aim);
    }
    
    public void removePlace(Place place) {
        places.remove(place.id());
        inputs.remove(place);
        
        for (List<Place> p : outputs.values())
            p.remove(place);
        
        Iterator<Pair<Place, Transition>> inarciter =
            inarcs.keySet().iterator();
        while (inarciter.hasNext()) {
            if (inarciter.next().first().equals(place))
                inarciter.remove();
        }
        
        Iterator<Pair<Transition, Place>> outarciter =
            outarcs.keySet().iterator();
        while (outarciter.hasNext()) {
            if (outarciter.next().second().equals(place))
                outarciter.remove();
        }

        marking.remove(place);
        preTransitions.remove(place);
        
        for (List<Place> p : prePlaces.values())
            p.remove(place);
    }
    
    public void removeTransition(Transition transition) {
        transitions.remove(transition.id());
        outputs.remove(transition);
        
        for (List<Transition> t : inputs.values())
            t.remove(transition);

        Iterator<Pair<Place, Transition>> inarciter = inarcs.keySet().iterator();
        while (inarciter.hasNext()) {
            if (inarciter.next().second().equals(transition))
                inarciter.remove();
        }
        
        Iterator<Pair<Transition, Place>> outarciter = outarcs.keySet().iterator();
        while (outarciter.hasNext()) {
            if (outarciter.next().first().equals(transition))
                outarciter.remove();
        }
        
        prePlaces.remove(transition);
        
        for (List<Transition> t : preTransitions.values())
            t.remove(transition);
    }
    
    /**
     * Returns the token count associated with a place. Default is 0.
     * @param place The place.
     * @return A nonnegative integer.
     */
    public Long getTokens(Place place) {
        return marking.get(place);
    }
    
    /**
     * Change the token count of a place in the Petri net. 
     * @param place The place.
     * @param tokens A nonnegative integer indicating the new token count.
     */
    public void setTokens(Place place, Long tokens) {
        marking.put(place, tokens);
    }
    
    /**
     * Returns all places in the Petri net.
     * @return A read-only list of places.
     */
    public Collection<Place> places() {
        return Collections.unmodifiableCollection(places.values());
    }
    
    /**
     * Returns all transitions in the Petri net.
     * @return A read-only list of transitions.
     */
    public Collection<Transition> transitions() {
        return Collections.unmodifiableCollection(transitions.values());
    }
    
    /**
     * Searches the Petri net for a place with the given ID.
     * @param id The ID to search for.
     * @return Returns the {@link Place} instance if the ID was found,
     *          <code>null</code> otherwise.
     */
    public Place findPlace(int id) {
        return places.get(id);
    }
    
    /**
     * Searches the Petri net for a transition with the given ID.
     * @param id The ID to search for.
     * @return Returns the {@link Transition} instance if the ID was found,
     *          <code>null</code> otherwise.
     */
    public Transition findTransition(int id) {
        return transitions.get(id);
    }
    
    /**
     * Returns the (initial) marking of the Petri net.
     * @return A map that associates an integer counter with each place to
     *          denote its token count. Each place in the Petri net has an
     *          associated counter, even if it is 0.
     */
    public Map<Place, Long> marking() {
        return Collections.unmodifiableMap(marking);
    }
    
    /**
     * Returns all transitions directly reachable from place {@code from},
     * i.e. all transitions to which {code from} is an <em>input place</em>.
     * @param from The starting place.
     * @return A read-only list of transitions.
     */
    public List<Transition> getTransitionsFor(Place from) {
        return Collections.unmodifiableList(inputs.get(from));
    }
    
    /**
     * Returns all <em>output places</em> of transition {code from}. 
     * @param from The starting transition.
     * @return A read-only list of places.
     */
    public List<Place> getPlacesFor(Transition from) {
        return Collections.unmodifiableList(outputs.get(from));
    }
    
    /**
     * Returns all transitions for which {@code to} is an output place.
     * @param to The output place.
     * @return A read-only list of transitions.
     */
    public List<Transition> getInputTransitionsFor(Place to) {
        return Collections.unmodifiableList(preTransitions.get(to));
    }
    
    /**
     * Returns all input places to transition {@code to}.
     * @param to The output transition.
     * @return A read-only list of places.
     */
    public List<Place> getInputPlacesFor(Transition to) {
        return Collections.unmodifiableList(prePlaces.get(to));
    }
    
    /**
     * Retrieves an input arc, identified by the associated place and transition.
     * @param from A place.
     * @param to A transition.
     * @return The arc from {@code from} to {@code to}.
     */
    public Arc getArc(Place from, Transition to) {
        return inarcs.get(Pair.of(from, to));
    }

     /**
     * Retrieves the number of transitions of the petrinet.
     * @return Number of transtiotions.
     */
    public int getNbrOfTransitions() {
        return transitions.size();
    }

         /**
     * Retrieves the number of places of the petrinet.
     * @return Number of places.
     */
    public int getNbrOfPlaces() {
        return places.size();
    }

    /**
     * Retrieves an output arc, identified by the associated transition and place.
     * @param from A transition.
     * @param to A place.
     * @return The arc from {@code from} to {@code to}.
     */
    public Arc getArc(Transition from, Place to) {
        return outarcs.get(Pair.of(from, to));
    }

    @SuppressWarnings("unchecked")
    @Override
    public void readExternal(ObjectInput in) throws IOException,
            ClassNotFoundException {
        super.readExternal(in);
        places = (Map<Integer, Place>) in.readObject();
        transitions = (Map<Integer, Transition>) in.readObject();
        inputs = (Map<Place, List<Transition>>) in.readObject();
        outputs = (Map<Transition, List<Place>>) in.readObject();
        inarcs = (Map<Pair<Place, Transition>, Arc>) in.readObject();
        outarcs = (Map<Pair<Transition, Place>, Arc>) in.readObject();
        marking = (Map<Place, Long>) in.readObject();
        
        // Revive all Petri net components by setting their container.
        // At the same time, initialize the inverse lookup tables to save one
        // loop.
        
        for (Place p : places()) {
            p.setContainer(this);
            preTransitions.put(p, new ArrayList<Transition>());
        }
        for (Transition t : transitions()) {
            t.setContainer(this);
            prePlaces.put(t, new ArrayList<Place>());
        }
        
        // Build up inverse lookup tables.
        // CONSIDER: This may be slow.
        
        for (Map.Entry<Place, List<Transition>> entry : inputs.entrySet())
            for (Transition transition : entry.getValue())
                prePlaces.get(transition).add(entry.getKey());
        
        for (Map.Entry<Transition, List<Place>> entry : outputs.entrySet())
            for (Place place : entry.getValue())
                preTransitions.get(place).add(entry.getKey());
    }

    public Map<Pair<Place, Transition>, Arc> getInarcs() {
        return inarcs;
    }

    public Map<Pair<Transition, Place>, Arc> getOutarcs() {
        return outarcs;
    }

    @Override
    public void writeExternal(ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeObject(places);
        out.writeObject(transitions);
        out.writeObject(inputs);
        out.writeObject(outputs);
        out.writeObject(inarcs);
        out.writeObject(outarcs);
        out.writeObject(marking);
    }
}
