/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.*;
import java.io.*;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import layout.TableLayout;
import monalisa.MonaLisa;
import monalisa.Project;
import monalisa.Settings;
import monalisa.ToolStatusUpdateEvent;
import monalisa.ToolStatusUpdateListener;
import monalisa.addons.netviewer.NetViewer;
import monalisa.addons.netviewer.View;
import monalisa.addons.util.MyOutputFileFilter;
import monalisa.data.Pair;
import monalisa.data.input.InputHandler;
import monalisa.data.input.PetriNetInputHandlers;
import monalisa.data.input.TInputHandler;
import monalisa.data.input.TInputHandlers;
import monalisa.data.output.OutputHandler;
import monalisa.data.output.PetriNetOutputHandlers;
import monalisa.data.pn.PetriNet;
import monalisa.gui.components.SplashScreen;
import monalisa.gui.components.StatusBar;
import monalisa.resources.ResourceManager;
import monalisa.resources.StringResources;
import monalisa.results.Configuration;
import monalisa.results.Result;
import monalisa.tools.ErrorLog;
import monalisa.tools.Tool;
import monalisa.tools.mcts.MctsTool;
import monalisa.tools.pinv.PInvariantTool;
import monalisa.tools.tinv.TInvariantTool;
import monalisa.util.FileUtils;


@SuppressWarnings("serial")
public final class MainDialog extends JFrame implements ActionListener, HierarchyBoundsListener, ToolStatusUpdateListener {

    private static final String MENU_FILE_NEW_ACTION = "MENU_FILE_NEW_ACTION";
    private static final String MENU_FILE_OPEN_ACTION = "MENU_FILE_OPEN_ACTION";
    private static final String MENU_FILE_SAVE_ACTION = "MENU_FILE_SAVE_ACTION";
    private static final String MENU_FILE_EXPORT_ACTION = "MENU_FILE_EXPORT_ACTION";
    private static final String MENU_FILE_EXPORT_PETRINET_ACTION = "MENU_FILE_EXPORT_PETRINET_ACTION";
    private static final String MENU_TFILE_LOAD_ACTION = "MENU_TFILE_LOAD_ACTION";
    private static final String MENU_FILE_EXIT_ACTION = "MENU_FILE_EXIT_ACTION";
    private static final String MENU_PROJECT_RUN = "MENU_PROJECT_RUN";
    private static final String MENU_PROJECT_STOP = "MENU_PROJECT_STOP";
    private static final String MENU_SHOW_NV = "MENU_SHOW_NV";
    private static final String EMPTY_PROJECT = "EMPTY_PROJECT";
    private static final String MENU_FILE_SAVE_AS_ACTION = "MENU_FILE_SAVE_AS_ACTION";
    private static final String ABOUT_ACTION = "ABOUT_ACTION";
    private static final String HELP_ACTION = "HELP_ACTION";
    private static final String MENU_FILE_CONVERTER_ACTION = "MENU_FILE_CONVERTER_ACTION";

    private static final Dimension DEFAULT_DIMENSION = new Dimension(400, 500);

    private static final ResourceManager resources = ResourceManager.instance();
    private static final StringResources strings = resources.getDefaultStrings(); 

    // Addons
    private NetViewer netViewer;

    private JMenuBar menuBar;
    private JMenu mainMenu, projectMenu, helpMenu, menuFileOpenRecently;
    private JMenuItem menuFileNew, menuFileImport, menuFileOpen, menuFileSave, menuFileSaveAs,
                      menuTFileLoad, menuFileExportResults, menuFileExportPetriNet,
                      menuFileExit, menuProjectRun, menuHelpAbout, menuHelpHelp, menuConverter;

    private JButton fileNewButton, fileOpenButton, fileSaveButton, fileExportResultsButton,
                    fileExportPetriNetButton, projectRunButton, showNVButton, loadProject, emptyProject, newProject;

    private Icon fileNewIcon, fileOpenIcon, fileSaveIcon, fileExportResultsIcon,
                 fileExportPetriNetIcon, projectRunIcon, projectStopIcon, showNVIcon;

    private JToolBar toolBar;
    private JPanel mainContainer, contentPanel, buttonContainer;
    private JScrollPane scrollPane;
    private StatusBar statusBar;
    private SplashScreen splash;

    private String documentTitle;
    private Project project;

    public MainDialog() {
        addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent e) {
                     exitApplication();
            }
        });
        
        System.err.println("Using MonaLisa Version "+strings.get("CurrentVersion"));
        
        initComponent();
    }

    private void initComponent() {
        setDocumentTitle(null);
        setIconImage(resources.getImage("icon-16.png"));
        
        initIcons();
        initMenuBar();
        initToolBar();

        addKeyListener(new KeyListener() {
            private Integer keyCode;

            public void keyTyped(KeyEvent e) { }

            public void keyPressed(KeyEvent e) { }

            public void keyReleased(KeyEvent e) {
                keyCode = e.getKeyCode();

                if(keyCode.equals(KeyEvent.VK_S) && e.isControlDown()) {
                    try {
                        project.save();
                    } catch (IOException ex) {
                        Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            }

        });

        splash = new SplashScreen();
        splash.setPreferredSize(DEFAULT_DIMENSION);

        loadProject = new JButton();
        loadProject.setText(strings.get("LoadProject"));
        loadProject.setActionCommand(MENU_FILE_OPEN_ACTION);
        loadProject.addActionListener(this);

        emptyProject = new JButton();
        emptyProject.setText(strings.get("CreateEmptyProject"));
        emptyProject.setActionCommand(EMPTY_PROJECT);
        emptyProject.addActionListener(this);

        newProject = new JButton();
        newProject.setText(strings.get("CreateNewProject"));
        newProject.setActionCommand(MENU_FILE_NEW_ACTION);
        newProject.addActionListener(this);

        double size[][] =
        {{TableLayout.FILL, TableLayout.PREFERRED, TableLayout.FILL},
        {5, TableLayout.PREFERRED, 5, TableLayout.PREFERRED, 5, TableLayout.PREFERRED}};
        splash.setLayout (new TableLayout(size));
        splash.add(emptyProject ,"1,1");
        splash.add(newProject ,"1,3");
        splash.add(loadProject ,"1,5");

        contentPanel = new JPanel();
        contentPanel.setLayout(new GroupLayout(contentPanel));
        scrollPane =
            new JScrollPane(contentPanel,
                JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.setPreferredSize(DEFAULT_DIMENSION);
        scrollPane.setBorder(BorderFactory.createEmptyBorder());
        scrollPane.getViewport().addHierarchyBoundsListener(this);

        mainContainer = new JPanel();
        mainContainer.setLayout(new BoxLayout(mainContainer,  BoxLayout.PAGE_AXIS));
        mainContainer.add(Box.createRigidArea(new Dimension(0, 3)));
        mainContainer.add(splash);
       
        statusBar = new StatusBar();
        
        buttonContainer = new JPanel();
        buttonContainer.setLayout(new BorderLayout());
        buttonContainer.add(projectRunButton, BorderLayout.WEST);
        buttonContainer.add(fileExportResultsButton, BorderLayout.CENTER);
        buttonContainer.add(this.fileExportPetriNetButton, BorderLayout.EAST);
        buttonContainer.add(statusBar, BorderLayout.SOUTH);
              
        Container contentPane = getContentPane();
        contentPane.setLayout(new BorderLayout());
        contentPane.add(toolBar, BorderLayout.NORTH);
        contentPane.add(mainContainer);
        contentPane.add(buttonContainer, BorderLayout.SOUTH);
        
        pack();
        projectClosed();

        setDefaultCloseOperation(EXIT_ON_CLOSE);
        setLocationRelativeTo(null);
        setVisible(true);
    }
    
    @Override
    public void ancestorResized(HierarchyEvent e) {
        if (e.getChanged() == scrollPane) {
            for (Component comp : contentPanel.getComponents()) {
                Dimension newSize = new Dimension(scrollPane.getViewport().getWidth(), comp.getHeight());
                comp.setSize(newSize);
                comp.setPreferredSize(newSize);
            }
        }
    }
    
    // Not needed.
    @Override
    public void ancestorMoved(HierarchyEvent e) { }

    private void initIcons() {
        fileNewIcon = resources.getIcon("new_project.png");
        fileOpenIcon = resources.getIcon("open_project.png");
        fileSaveIcon = resources.getIcon("save_project.png");
        fileExportResultsIcon = resources.getIcon("export_results.png");
        fileExportPetriNetIcon = resources.getIcon("export.png");
        projectRunIcon = resources.getIcon("run_tools.png");
        projectStopIcon = resources.getIcon("stop_tools.png");
        showNVIcon = resources.getIcon("netviewer.png");
    }

    private void initMenuBar() {
        menuFileNew = new JMenuItem(strings.get("FileNew"), fileNewIcon);
        menuFileNew.setActionCommand(EMPTY_PROJECT);
        menuFileNew.addActionListener(this);
        
        menuFileImport = new JMenuItem(strings.get("CreateNewProject"), resources.getIcon("import.png"));
        menuFileImport.setActionCommand(MENU_FILE_NEW_ACTION);
        menuFileImport.addActionListener(this);

        menuFileOpen = new JMenuItem(strings.get("FileOpen"), fileOpenIcon);
        menuFileOpen.setActionCommand(MENU_FILE_OPEN_ACTION);
        menuFileOpen.addActionListener(this);

        menuFileOpenRecently = new JMenu(strings.get("FileOpenRecently"));
        menuFileOpenRecently.setIcon(fileOpenIcon); 
        
        String recentlyProjects = Settings.get("recentlyProjects");
        
        if(!recentlyProjects.isEmpty()) {
            String projects[] = recentlyProjects.split(",");
            
            for(int i = projects.length-1; i >= 0; i--) {
                JMenuItem item = new JMenuItem();

                if(projects[i].length() > 40 ) {
                    String subString = projects[i].substring(projects[i].lastIndexOf(System.getProperty("file.separator"))+1, projects[i].length());                
                     if(subString.length() <= 40) {
                        item.setText(".../"+subString);
                    } else {
                        item.setText("..."+projects[i].substring(projects[i].length()-31, projects[i].length()));  
                    }
                } else {
                   item.setText(projects[i]);
                }

                item.setToolTipText(projects[i]);                       

                item.addActionListener(new ActionListener() {
                     public void actionPerformed(ActionEvent e) {
                        File file = new File(((JMenuItem)e.getSource()).getToolTipText());
                        if(file.exists()) {                         
                            try {
                                openProject(file);
                            } catch (IOException ex) {
                                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                            } catch (ClassNotFoundException ex) {
                                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                            } catch (InterruptedException ex) {
                                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                            }
                        } else {
                            JOptionPane.showMessageDialog(((JMenuItem)e.getSource()).getRootPane(), "This project file is not existing");
                        }
                     }                   
                });

                menuFileOpenRecently.add(item);
            }
        }        
        
        menuConverter = new JMenuItem(strings.get("FileConverter"), resources.getIcon("open_project.png"));
        menuConverter.setActionCommand(MENU_FILE_CONVERTER_ACTION);
        menuConverter.addActionListener(this);        
        
        menuFileSave = new JMenuItem(strings.get("FileSave"), fileSaveIcon);
        menuFileSave.setActionCommand(MENU_FILE_SAVE_ACTION);
        menuFileSave.addActionListener(this);

        menuFileSaveAs = new JMenuItem(strings.get("FileSaveAs"), fileSaveIcon);
        menuFileSaveAs.setActionCommand(MENU_FILE_SAVE_AS_ACTION);
        menuFileSaveAs.addActionListener(this);

        menuTFileLoad = new JMenuItem(strings.get("TFileLoad"), fileOpenIcon);
        menuTFileLoad.setActionCommand(MENU_TFILE_LOAD_ACTION);
        menuTFileLoad.addActionListener(this);

        menuFileExportResults = new JMenuItem(strings.get("FileExport"), fileExportResultsIcon);
        menuFileExportResults.setActionCommand(MENU_FILE_EXPORT_ACTION);
        menuFileExportResults.addActionListener(this);

        menuFileExportPetriNet = new JMenuItem(strings.get("ExportPetriNet"), fileExportPetriNetIcon);
        menuFileExportPetriNet.setActionCommand(MENU_FILE_EXPORT_PETRINET_ACTION);
        menuFileExportPetriNet.addActionListener(this);
        
        menuFileExit = new JMenuItem(strings.get("FileExit"), resources.getIcon("delete.png"));
        menuFileExit.setActionCommand(MENU_FILE_EXIT_ACTION);
        menuFileExit.addActionListener(this);

        mainMenu = new JMenu(strings.get("File"));
        mainMenu.add(menuFileNew);
        mainMenu.add(menuFileImport);
        mainMenu.add(menuFileOpen);
        mainMenu.add(menuFileOpenRecently);
        mainMenu.add(menuConverter);
        //mainMenu.add(menuTFileLoad);
        mainMenu.add(menuFileSave);
        mainMenu.add(menuFileSaveAs);
        mainMenu.addSeparator();
        mainMenu.add(menuFileExportPetriNet);
        mainMenu.add(menuFileExportResults);
        mainMenu.addSeparator();
        mainMenu.add(menuFileExit);

        menuProjectRun = new JMenuItem(strings.get("ProjectRun"), projectRunIcon);
        menuProjectRun.setActionCommand(MENU_PROJECT_RUN);
        menuProjectRun.addActionListener(this);

        projectMenu = new JMenu(strings.get("Project"));
        projectMenu.add(menuProjectRun);

        menuHelpHelp = new JMenuItem(strings.get("Help"), resources.getIcon("help.png"));
        menuHelpHelp.setActionCommand(HELP_ACTION);
        menuHelpHelp.addActionListener(this);

        menuHelpAbout = new JMenuItem(strings.get("About"), resources.getIcon("about.png"));
        menuHelpAbout.setActionCommand(ABOUT_ACTION);
        menuHelpAbout.addActionListener(this);

        helpMenu = new JMenu(strings.get("Help"));
        helpMenu.add(menuHelpHelp);
        helpMenu.add(menuHelpAbout);

        menuBar = new JMenuBar();
        menuBar.add(mainMenu);
        menuBar.add(projectMenu);
        menuBar.add(helpMenu);

        setJMenuBar(menuBar);
    }

    private void initToolBar() {
        fileNewButton = new JButton(fileNewIcon);
        fileNewButton.setToolTipText(strings.get("FileNew"));
        fileNewButton.setActionCommand(EMPTY_PROJECT);
        fileNewButton.addActionListener(this);

        fileOpenButton = new JButton(fileOpenIcon);
        fileOpenButton.setToolTipText(strings.get("FileOpen"));
        fileOpenButton.setActionCommand(MENU_FILE_OPEN_ACTION);
        fileOpenButton.addActionListener(this);

        fileSaveButton = new JButton(fileSaveIcon);
        fileSaveButton.setToolTipText(strings.get("FileSave"));
        fileSaveButton.setActionCommand(MENU_FILE_SAVE_ACTION);
        fileSaveButton.addActionListener(this);

        fileExportResultsButton = new JButton(strings.get("FileExport"), fileExportResultsIcon);
        fileExportResultsButton.setToolTipText(strings.get("FileExportTT"));
        fileExportResultsButton.setActionCommand(MENU_FILE_EXPORT_ACTION);
        fileExportResultsButton.addActionListener(this);

        fileExportPetriNetButton = new JButton(strings.get("ExportPetriNet"), fileExportPetriNetIcon);
        fileExportPetriNetButton.setToolTipText(strings.get("ExportPetriNetTT"));
        fileExportPetriNetButton.setActionCommand(MENU_FILE_EXPORT_PETRINET_ACTION);
        fileExportPetriNetButton.addActionListener(this);

        projectRunButton = new JButton(strings.get("ProjectRun"), projectRunIcon);
        projectRunButton.setToolTipText(strings.get("ProjectRun"));
        projectRunButton.setActionCommand(MENU_PROJECT_RUN);
        projectRunButton.addActionListener(this);

        showNVButton = new JButton(showNVIcon);
        showNVButton.setToolTipText(strings.get("NVButton"));
        showNVButton.setActionCommand(MENU_SHOW_NV);
        showNVButton.addActionListener(this);

        toolBar = new JToolBar();
        toolBar.setFloatable(false);
        toolBar.setRollover(true);
        toolBar.add(fileNewButton);
        toolBar.add(fileOpenButton);
        toolBar.add(fileSaveButton);
//        toolBar.add(fileExportResultsButton);
//        toolBar.add(fileExportPetriNetButton);
        toolBar.addSeparator();
//        toolBar.add(projectRunButton);
        toolBar.addSeparator();
        toolBar.add(showNVButton);
        toolBar.add(Box.createHorizontalGlue());
    }

    public String getDocumentTitle() {
        return documentTitle;
    }

    public void setDocumentTitle(String documentTitle) {
        this.documentTitle = documentTitle;
        String doc = getDocumentTitle() == null ? strings.get("NoProject") : getDocumentTitle().equals("") ? strings.get("EmptyProject") : getDocumentTitle();
        setTitle(MonaLisa.APPLICATION_TITLE + " – " + doc);
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        final String action = e.getActionCommand();

        if(action == MENU_FILE_EXIT_ACTION)
            exitApplication();
        else if(action == MENU_FILE_NEW_ACTION) {
            try {
                createNewProject();
            } catch (InterruptedException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if(action == MENU_FILE_OPEN_ACTION) {
            try {
                openProject();
            } catch (InterruptedException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if(action == MENU_FILE_SAVE_ACTION)
            try {
                saveProject();
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        else if(action == MENU_FILE_SAVE_AS_ACTION)
            try {
                if(project.getPath() != null);
                    saveProject();
                createNewProject(project.getPetriNet());
            } catch (IOException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (InterruptedException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        else if(action == MENU_FILE_EXPORT_ACTION)
            exportResults();
        else if(action == MENU_FILE_EXPORT_PETRINET_ACTION)
            try {
                exportPetriNet();
            } catch (IOException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        else if(action == MENU_PROJECT_RUN)
            runSelectedTools();
        else if(action == MENU_PROJECT_STOP)
            stopRunningTools();
        else if(action == MENU_SHOW_NV)
            showNV();
        else if(action == EMPTY_PROJECT) {
            project = null;
            PetriNet pn = new PetriNet();
            try {
                createNewProject(pn);
            } catch (InterruptedException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (ClassNotFoundException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if (action == MENU_TFILE_LOAD_ACTION) {
            try {
                loadTFile();
            } catch (InterruptedException ex) {
                Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        else if(action == HELP_ACTION) {
            HelpDialog hd = new HelpDialog();
            hd.setVisible(true);
        }
        else if(action == ABOUT_ACTION) {
            AboutDialog ad = new AboutDialog();
            ad.setVisible(true);
        }
        else if(action == MENU_FILE_CONVERTER_ACTION) {
            converter();
        }        
    }

    private void converter() {
        ConverterDialog dialog = new ConverterDialog(MonaLisa.appMainWindow());
    }    
    
    private void runSelectedTools() {
        if(!this.project.getPetriNet().places().isEmpty() && !this.project.getPetriNet().transitions().isEmpty()) {          
            setRunButtons(MENU_PROJECT_STOP, strings.get("StopRunningTools"), projectStopIcon);
            statusBar.setProgressValue(0);
            statusBar.setIndeterminateProgress(true);
            statusBar.setProgressVisible(true);
            project.runSelectedTools();
        }
    }
    
    private void stopRunningTools() {
        project.stopRunningTools();
        restoreRunButtons();
    }
    
    private void restoreRunButtons() {
        setRunButtons(MENU_PROJECT_RUN, strings.get("ProjectRun"), projectRunIcon);
        statusBar.setProgressVisible(false);
    }
    
    private void setRunButtons(String action, String text, Icon icon) {
        menuProjectRun.setIcon(icon);
        menuProjectRun.setText(text);
        menuProjectRun.setActionCommand(action);
        projectRunButton.setIcon(icon);
        projectRunButton.setToolTipText(text);
        projectRunButton.setActionCommand(action);
    }

    private void exitApplication() {
        if(this.project != null) {
            if(this.project.isProjectChanged()) {        
                JOptionPane optionPane = new JOptionPane("Save the current Project before closing?", JOptionPane.QUESTION_MESSAGE, JOptionPane.YES_NO_OPTION);
                JDialog dialog = optionPane.createDialog(this, "Wait a moment...");
                dialog.setVisible(true);
                Integer selectedValue = (Integer) optionPane.getValue();

                if(selectedValue == JOptionPane.YES_OPTION) {
                    try {
                    netViewer.saveView(true, false);
                    } catch (ClassNotFoundException ex) {
                        Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                    } catch (IOException ex) {
                        Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                    }

                    System.exit(0);
                }
                else if(selectedValue == JOptionPane.NO_OPTION) {
                    System.exit(0);
                } 
            }
            else {
                System.exit(0);
            }  
        }
        System.exit(0);
    }

    /***
     * Creates a new project from a choosen Petri net file
     * @throws IOException
     * @throws ClassNotFoundException
     * @throws InterruptedException
     */
    private void createNewProject() throws IOException, ClassNotFoundException, InterruptedException {
        saveProject();

//        // Project file
//        File projectFile;
//        JFileChooser projectLocationChooser = new JFileChooser();
//        projectLocationChooser.setFileFilter(new FileFilter() {
//            public boolean accept(File f) {
//                return f.isDirectory() || Project.FILENAME_EXTENSION.equalsIgnoreCase(FileUtils.getExtension(f));
//            }
//
//            public String getDescription() {
//                return strings.get("ProjectFileType");
//            }
//        });
//
//        projectLocationChooser.setDialogTitle(strings.get("NewProjectLocation"));
//        if (projectLocationChooser.showDialog(this, strings.get("ChooseProjectLocationButton")) != JFileChooser.APPROVE_OPTION)
//            return;
//
//        projectFile = projectLocationChooser.getSelectedFile();

//        if (!Project.FILENAME_EXTENSION.equalsIgnoreCase(FileUtils.getExtension(projectFile)))
//            projectFile = new File(projectFile.getAbsolutePath() + "." + Project.FILENAME_EXTENSION);
//          //projectFile = new File(petriNetFile.getAbsolutePath().substring(0, petriNetFile.getAbsolutePath().lastIndexOf(".")) + "." + Project.FILENAME_EXTENSION);

        // PN File
        String latestDirectory = Settings.get("latestDirectory");
        
        JFileChooser petriNetFileChooser;
        if(latestDirectory.isEmpty()) {
            petriNetFileChooser = new JFileChooser();
        } else {
            petriNetFileChooser = new JFileChooser(new File(latestDirectory));
        }
        
        for (final InputHandler handler : PetriNetInputHandlers.getHandlers()) {
            petriNetFileChooser.addChoosableFileFilter(new FileFilter() {
                public String getDescription() {
                    return strings.get(handler.getClass().getSimpleName()
                        .replaceAll("InputHandler", "FileType"));
                }
                public boolean accept(File f) {
                    try {
                        return f.isDirectory() || handler.isKnownFile(f);
                    } catch (IOException e) {
                        return false;
                    }
                }
            });
        }

        petriNetFileChooser.setFileFilter(new FileFilter() {
            public String getDescription() {
                return strings.get("PetriNetFileType");
            }

            public boolean accept(File f) {
                try {
                    return f.isDirectory() || PetriNetInputHandlers.isKnownFile(f);
                } catch (IOException e) {
                    return false;
                }
            }
        });

        petriNetFileChooser.setDialogTitle(strings.get("ChooseAPetriNetFile"));
        if (petriNetFileChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
            return;

        File petriNetFile = petriNetFileChooser.getSelectedFile();
        Settings.set("latestDirectory", petriNetFile.getAbsolutePath().substring(0, petriNetFile.getAbsolutePath().lastIndexOf(System.getProperty("file.separator"))));

        try {
            project = Project.create(petriNetFile);
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(this, strings.get("ErrorReadingFileMessage", petriNetFile), strings.get("ErrorReadingFileTitle"), JOptionPane.ERROR_MESSAGE);
            return;
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this, strings.get("ErrorReadingPetriNetFileMessage", petriNetFile), strings.get("ErrorReadingFileTitle"), JOptionPane.ERROR_MESSAGE);
            return;
        }

//        try {
//            project.save(projectFile);
//        } catch (IOException ex) {
//            JOptionPane.showMessageDialog(this, strings.get("ErrorWritingFileMessage"), strings.get("ErrorWritingFileTitle"), JOptionPane.ERROR_MESSAGE);
//            ex.printStackTrace();
//            return;
//        }

        projectLoaded();
    }

    /**
     * Creates a new project out of a existing project
     * @param pn
     * @throws IOException
     * @throws ClassNotFoundException
     * @throws InterruptedException
     */
    public void createNewProject(PetriNet pn) throws IOException, ClassNotFoundException, InterruptedException {
        File projectFile = null;

        if(project != null) {
            JFileChooser projectLocationChooser = new JFileChooser();
                    projectLocationChooser.setFileFilter(new FileFilter() {
                        public boolean accept(File f) {
                            return f.isDirectory()
                                || Project.FILENAME_EXTENSION.equalsIgnoreCase(
                                    FileUtils.getExtension(f));
                        }

                        public String getDescription() {
                            return strings.get("ProjectFileType");
                        }
                    });

            projectLocationChooser.setDialogTitle(strings.get("NewProjectLocation"));
            if (projectLocationChooser.showDialog(this, strings.get("ChooseProjectLocationButton")) != JFileChooser.APPROVE_OPTION)
                return;

            projectFile = projectLocationChooser.getSelectedFile();

            if (!Project.FILENAME_EXTENSION.equalsIgnoreCase(FileUtils.getExtension(projectFile)))
                projectFile = new File(projectFile.getAbsolutePath() + "." + Project.FILENAME_EXTENSION);
        }
        
        if(project != null) {
            project = Project.create(pn, project);
            project.save(projectFile);
        }
        else 
            project = Project.create(pn);

        projectLoaded();
    }

    private void loadTFile() throws InterruptedException {
        JFileChooser TfileLocationChooser = new JFileChooser();

        for (final TInputHandler handler : TInputHandlers.getHandlers()) {
            TfileLocationChooser.addChoosableFileFilter(new FileFilter() {
                public String getDescription() {
                    return strings.get(handler.getClass().getSimpleName()
                        .replaceAll("TInputHandler", "FileType"));
                }

                public boolean accept(File f) {
                    try {
                        return f.isDirectory() || handler.isKnownFile(f);
                    } catch (IOException e) {
                        return false;
                    }
                }
            });
        }

        TfileLocationChooser.setDialogTitle(strings.get("LoadTFile"));
        if (TfileLocationChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
            return;

        File TFile = TfileLocationChooser.getSelectedFile();

        try {
            project.loadTFile(TFile);
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(this, strings.get("ErrorReadingFileMessage", TFile), strings.get("ErrorReadingFileTitle"), JOptionPane.ERROR_MESSAGE);
        }
    }

    private void openProject() throws IOException, ClassNotFoundException, InterruptedException {
        String latestDirectory = Settings.get("latestDirectory");
        
        JFileChooser projectLocationChooser;
        if(latestDirectory.isEmpty()) {
            projectLocationChooser = new JFileChooser();
        } else {
            projectLocationChooser = new JFileChooser(new File(latestDirectory));
        }
        
        projectLocationChooser.setFileFilter(new FileFilter() {
            public boolean accept(File f) {
                return f.isDirectory()
                    || Project.FILENAME_EXTENSION.equalsIgnoreCase(
                        FileUtils.getExtension(f));
            }

            public String getDescription() {
                return strings.get("ProjectFileType");
            }
        });

        projectLocationChooser.setDialogTitle(strings.get("LoadProject"));
        if (projectLocationChooser.showOpenDialog(this) != JFileChooser.APPROVE_OPTION)
            return;

        File projectFile = projectLocationChooser.getSelectedFile();
        Settings.set("latestDirectory", projectFile.getAbsolutePath().substring(0, projectFile.getAbsolutePath().lastIndexOf(System.getProperty("file.separator"))));        
        openProject(projectFile);
    }

    public void openProject(File projectFile) throws IOException, ClassNotFoundException, InterruptedException {
        try {
            project = Project.load(projectFile);
        } catch (IOException ex) {
            JOptionPane.showMessageDialog(this, strings.get("ErrorReadingFileMessage", projectFile), strings.get("ErrorReadingFileTitle"), JOptionPane.ERROR_MESSAGE);
            return;
        }

        String recentlyProjects = Settings.get("recentlyProjects");
        
        // If project are in the list, set it to the last place
        if(recentlyProjects.contains(projectFile.getAbsolutePath())) {
            recentlyProjects = recentlyProjects.replace(projectFile.getAbsolutePath()+",", "");      
        }
        else {       
            // If not, check if the list is to large and delte the first element
            String projects[] = recentlyProjects.split(",");
            if(projects.length == 10) {            
                String tmp = "";             
                for(int i = 1; i < 10; i++) {
                    tmp += projects[i] + ",";
                }
                recentlyProjects = tmp;
            }
        }   
        // Add the new project at the last place
        recentlyProjects += projectFile.getAbsolutePath()+",";
        
        Settings.set("recentlyProjects", recentlyProjects);
        Settings.writeToFile(Settings.getConfigFile());        
        
        projectLoaded();
    }

    private void saveProject() throws ClassNotFoundException {
        if(project != null) {
            try {
                if(netViewer != null)
                    netViewer.updatePetriNet();
                project.save();
            } catch (IOException ex) {
                JOptionPane.showMessageDialog(this, strings.get("ErrorWritingFileMessage"), strings.get("ErrorWritingFileTitle"), JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    public void exportPetriNet() throws IOException {
        JFileChooser petriNetFileChooser = new JFileChooser(project.getPath());
        petriNetFileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
        petriNetFileChooser.setDialogTitle(strings.get("ChooseAPetriNetFileName"));
        petriNetFileChooser.setAcceptAllFileFilterUsed(false);

        for (final OutputHandler handler : PetriNetOutputHandlers.getHandlers()) {
            petriNetFileChooser.addChoosableFileFilter(new MyOutputFileFilter(handler));
        }

        if (petriNetFileChooser.showDialog(this, "Export") != JFileChooser.APPROVE_OPTION)
            return;

        File petriNetFile = petriNetFileChooser.getSelectedFile();
        MyOutputFileFilter selectedFileFilter = ((MyOutputFileFilter)petriNetFileChooser.getFileFilter());
        petriNetFile = selectedFileFilter.checkFileNameForExtension(petriNetFile);
        selectedFileFilter.getHandler().save(new FileOutputStream(petriNetFile), project.getPetriNet());
    }

    private void exportResults() {
        ExportDialog exportDialog = new ExportDialog(this, project);

        if(exportDialog.isCancelled())
            return;

        Map<Pair<Class<? extends Tool>,Configuration>,String> exports = exportDialog.exportPaths();
        boolean retainAll = false;
        boolean deleteAll = false;
        
        for (Pair<Class<? extends Tool>, Configuration> export : exports.keySet()) {
            // Determine output file name.
            Result result = project.getResult(export.first(), export.second());
            File outputFile = new File(exportDialog.exportPaths().get(export));
            
            // See whether file already exists.
            if (outputFile.exists()) {
                if (retainAll)
                    continue;
                else if (deleteAll) {
                    // Do nothing.
                }
                else {
                    ConfirmOverwriteDialog dialog = new ConfirmOverwriteDialog(
                        this, outputFile.getName());
                    if (dialog.getResult() == ConfirmOverwriteDialog.YES) {
                        // Do nothing.
                    }
                    else if (dialog.getResult() == ConfirmOverwriteDialog.YES_FOR_ALL)
                        deleteAll = true;
                    else if (dialog.getResult() == ConfirmOverwriteDialog.NO)
                        continue;
                    else if (dialog.getResult() == ConfirmOverwriteDialog.NO_FOR_ALL) {
                        retainAll = true;
                        continue;
                    }
                }
            }
            
            try {
                result.export(outputFile, export.second(), project);
            } catch (IOException ex) {
                // TODO Proper error handling.
                System.out.printf("Unable to write output file %s\n", outputFile);
            }
        }
    }
    
    public void projectLoaded() throws ClassNotFoundException, IOException, InterruptedException {
        // Unload previous project, if any.
        projectClosed();
        mainContainer.removeAll();
        mainContainer.add(scrollPane);
        project.addToolStatusUpdateListener(this);
        setDocumentTitle(project.getName());
        setProjectRelatedEnabled(true);
        project.createUI(contentPanel, strings);
        recursivelyDoLayout(mainContainer);
        menuTFileLoad.setEnabled(true);

        if(project.getPath() == null)
            menuFileSaveAs.setEnabled(false);
        else
            menuFileSaveAs.setEnabled(true);

        if(netViewer != null) {
            netViewer.closeAllFrames();
            netViewer.dispose();
        }
        netViewer = new NetViewer(this, project);
    }

    private void projectClosed() {
        setDocumentTitle(null);
        setProjectRelatedEnabled(false);
        contentPanel.removeAll();
        contentPanel.doLayout();
    }
    
    private void setProjectRelatedEnabled(boolean enabled) {
        menuFileSave.setEnabled(enabled);
        menuFileSaveAs.setEnabled(enabled);
        fileSaveButton.setEnabled(enabled);
        menuFileExportResults.setEnabled(enabled);
        menuFileExportPetriNet.setEnabled(enabled);
        menuTFileLoad.setEnabled(enabled);
        fileExportResultsButton.setEnabled(enabled);
        fileExportPetriNetButton.setEnabled(enabled);
        projectMenu.setEnabled(enabled);
        projectRunButton.setEnabled(enabled);
        showNVButton.setEnabled(enabled);        
    }

    private void recursivelyDoLayout(JPanel panel) {
        panel.doLayout();
        for (Component child : panel.getComponents()) {
            if (child instanceof JPanel)
                recursivelyDoLayout((JPanel) child);
        }
    }

    @Override
    public void updated(final ToolStatusUpdateEvent e) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {                
                switch (e.getStatus()) {
                    case STARTED:
                        statusBar.setStatusText(e.getToolName() + " started");
                        statusBar.setProgressValue(0);
                        statusBar.setIndeterminateProgress(true);
                        break;
                    case FINISHED:
                        statusBar.setIndeterminateProgress(false);
                        statusBar.setProgressValue(100);
                        statusBar.setStatusText(" ");

                        // NetViewer
                        try {
                            updateNetViewer();
                        } catch (IOException ex) {
                            Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                        } catch (InterruptedException ex) {
                            Logger.getLogger(MainDialog.class.getName()).log(Level.SEVERE, null, ex);
                        }

                        break;
                    case ABORTED:
                        statusBar.setStatusText(" ");
                        restoreRunButtons();
                        ErrorLog interruptedMessages = project.getToolMessages();
                        interruptedMessages.log("#ToolsInterruptedByUser", ErrorLog.Severity.WARNING);
                        MessageDialog.show(MainDialog.this, interruptedMessages);
                        break;
                    case PROGRESS:
                        statusBar.setIndeterminateProgress(false);
                        statusBar.setProgressValue(e.getProgress());
                        break;
                    case FINISHED_ALL:
                        statusBar.setStatusText(" ");
                        restoreRunButtons();
                        ErrorLog messages = project.getToolMessages();
                        if (messages.has(ErrorLog.Severity.WARNING) || messages.has(ErrorLog.Severity.ERROR))
                            MessageDialog.show(MainDialog.this, messages);
                        else
                            JOptionPane.showMessageDialog(MainDialog.this,
                                strings.get("ToolCalculationFinishedMessage"),
                                strings.get("ToolCalculationFinishedTitle"),
                                JOptionPane.INFORMATION_MESSAGE);
                        break;
                    default:
                        break;
                }
            }
        });
    }

    private void showNV()  {
        netViewer.showMe();
    }

    private void updateNetViewer() throws InterruptedException, IOException {
        if(project.hasResults(TInvariantTool.class))
            netViewer.addTinvsToComboBox();
        if(project.hasResults(PInvariantTool.class))
            netViewer.addPinvsToComboBox();
        if(project.hasResults(MctsTool.class))
            netViewer.addMctsToComboBox();

        if(project.getViewWrapper().getViewCounter() > 0) {
            View view = netViewer.getSelectedView();

            // Write the results to a byte array
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(bos);
            oos.writeObject(project.getAllResults());
            oos.flush();
            oos.close();
            bos.close();
            byte [] resultsData = bos.toByteArray();
            view.setResults(resultsData);
        }

        netViewer.setNetChanged(false);
    }

    public void setExportEnabled(Boolean value) {
        menuFileExportPetriNet.setEnabled(value);
        fileExportPetriNetButton.setEnabled(value);
    }

    public void updateUI() {
        contentPanel.removeAll();
        project.createUI(contentPanel, strings);
    }
}
