/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.distance;

import java.util.List;

import monalisa.data.pn.TInvariant;
import monalisa.resources.ResourceManager;
import monalisa.resources.StringResources;

/**
 * @author Katja
 * @author Konrad Rudolph
 * @version 1.0.1
 */

/**
 * Enth�lt verschiedene Distanzfunktionen.
 */
public class DistanceFunctions {
    public static final String SUM_OF_DIFFERENCES = "SumOfDifferences";
    public static final String NUMBER_OF_DIFFERENT_OCCURANCES =
        "SumOfXorOfSupports";
    public static final String TANIMOTO = "Tanimoto";
    public static final String BB = "BraunBlanquet";
    public static final String SIMPLEMATCHING = "SimpleMatching";
    
    private static final StringResources strings =
        ResourceManager.instance().getDefaultStrings();

    /**
     * Im wesentlichen ein leerer Konstruktor - liegt ein Objekt dieser Klasse
     * an.
     */
    public DistanceFunctions() { }
    
    /**
     * Return a human-readable, localized name of the distance function.
     * @param distanceFunction The distance function.
     * @return A human-readable text representing the function's name.
     */
    public static String getName(String distanceFunction) {
        return strings.get("DistanceFunction" + distanceFunction);
    }

    /**
     * Gibt ein DistanceFunc-Object zur�ck. Die konkrete Funktion wird anhand
     * des Parameters <CODE>funktionsName</CODE> bestimmt.
     * 
     * @param funktionsName
     *            Name der Funktion
     * @return Funktionsobjekt
     */
    public DistanceFunction getDistanceFunc(String funktionsName) {
        if (funktionsName.equals(SUM_OF_DIFFERENCES)) {
            return new SumOfDifferencesFunc();
        } else if (funktionsName.equals(NUMBER_OF_DIFFERENT_OCCURANCES)) {
            return new NumberOfDifferencesFunc();
        } else if (funktionsName.equals(TANIMOTO)) {
            return new Tanimoto();
        } else if (funktionsName.equals(BB)) {
            return new BraunBlanquet();
        } else if (funktionsName.equals(SIMPLEMATCHING)) {
            return new SimpleMatching();
        }
        return new SumOfDifferencesFunc();
    }

    /**
     * Eine normierte Distancefunc-Implementierung.
     */
    public class SumOfDifferencesFunc implements DistanceFunction {
        /**
         * Die Vektoren der beiden TInvarianten werden voneinander subtrahiert.
         * �ber die Absolutwerte der einzelnen Elemente des Ergebnisses wird
         * addiert. Dieses Ergebnis wird zur�ckgegeben.
         * 
         * @param t1
         *            erste TInvariante
         * @param t2
         *            zweite TInvariante
         * @return Ergebnis
         */
        public float getDistanceValue(TInvariant t1, TInvariant t2) {
            List<Integer> vectorInv1 = t1.asVector();
            List<Integer> vectorInv2 = t2.asVector();
            float ret = 0;
            float resa = 0;
            float resb = 0;
            float resc = 0;
            float resc1 = 0;

            for (int i = 0; i < vectorInv1.size(); i++) {
                if (vectorInv1.get(i) > 0) {
                    resa += vectorInv1.get(i);
                }
                if (vectorInv2.get(i) > 0) {
                    resb += vectorInv2.get(i);
                }
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) > 0) {
                    if (vectorInv2.get(i) == vectorInv1.get(i)) {
                        resc += vectorInv2.get(i);
                    } else {
                        resc1 = Math.abs(vectorInv1.get(i) - vectorInv2.get(i));
                        if (vectorInv1.get(i) > vectorInv2.get(i)) {
                            resc1 = vectorInv1.get(i) - resc1;
                        } else {
                            resc1 = vectorInv2.get(i) - resc1;
                        }
                        resc += resc1;
                    }
                }

                ret += Math.abs(vectorInv1.get(i) - vectorInv2.get(i));
            }

            return ((ret / (resa + resb - resc)) * 60);
        }
    }

    /**
     * Eine normierte Distancefunc-Implementierung.
     */
    public class NumberOfDifferencesFunc implements DistanceFunction {
        /**
         * Implementierung einer Distanzfunktion. Die Supportvektoren der beiden
         * Invarianten werden verglichen: F�r jeden Unterschied wird eins auf
         * das Ergebnis aufaddiert. Man k�nnte auch sagen: die beiden
         * Supportvektoren werden mit xor verkn�ft, die Elemente des Ergebnisses
         * werden aufaddiert. beispiel: distance( (0,4,0,1), (1,6,0,0) ) =
         * 1+0+0+1 =2
         * 
         * @param t1
         *            erste TInvariante
         * @param t2
         *            zweite TInvariante
         * @return Ergebnis
         */
        public float getDistanceValue(TInvariant t1, TInvariant t2) {
            List<Integer> vectorInv1 = t1.asVector();
            List<Integer> vectorInv2 = t2.asVector();

            float ret = 0;
            float resa = 0;
            float resb = 0;
            float resc = 0;
            for (int i = 0; i < vectorInv1.size(); i++) {
                int a = vectorInv1.get(i) > 0 ? 1 : 0;
                int b = vectorInv2.get(i) > 0 ? 1 : 0;
                if (vectorInv1.get(i) > 0)
                    resa += 1;
                if (vectorInv2.get(i) > 0)
                    resb += 1;
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) > 0)
                    resc += 1;
                if (a != b)
                    ret += 1;
            }
            return (ret / (resa + resb - resc)) * 30;
        }
    }

    public class Tanimoto implements DistanceFunction {
        public float getDistanceValue(TInvariant t1, TInvariant t2) {
            List<Integer> vectorInv1 = t1.asVector();
            List<Integer> vectorInv2 = t2.asVector();

            float resa = 0; // bei beiden Inv ist das Merkmal vorhanden
            float resb = 0; // nur bei Inv1 ist das Merkmal vorhanden
            float resc = 0; // nur bei Inv2 ist das Merkmal vorhanden
            float res; // result

            for (int i = 0; i < vectorInv1.size(); i++) {
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) > 0)
                    resa += 1;
                if (vectorInv1.get(i) > 0 && vectorInv2.get(i) == 0)
                    resb += 1;
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) == 0)
                    resc += 1;
            }

            // Umwandlung similarity in distance
            res = resa / (resa + resb + resc);
            res = 1 - res;
            return res * 30;
        }
    }

    public class BraunBlanquet implements DistanceFunction {
        public float getDistanceValue(TInvariant t1, TInvariant t2) {
            List<Integer> vectorInv1 = t1.asVector();
            List<Integer> vectorInv2 = t2.asVector();

            float resa = 0; // bei beiden Inv ist das Merkmal vorhanden
            float resb = 0; // bei Inv1 ist das Merkmal vorhanden
            float resc = 0; // bei Inv2 ist das Merkmal vorhanden
            float nenner = 0;
            float res; // result

            for (int i = 0; i < vectorInv1.size(); i++) {
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) > 0)
                    resa += 1;
                if (vectorInv1.get(i) > 0)
                    resb += 1;
                if (vectorInv2.get(i) > 0)
                    resc += 1;
            }
            nenner = Math.max(resb, resc);

            // Umwandlung similarity in distance
            res = resa / nenner;
            res = 1 - res;
            System.out.println("a, b, c, nenner: " + resa + " " + resb + " "
                    + resc + " " + nenner);
            return res * 30;
        }
    }

    public class SimpleMatching implements DistanceFunction {
        public float getDistanceValue(TInvariant t1, TInvariant t2) {
            List<Integer> vectorInv1 = t1.asVector();
            List<Integer> vectorInv2 = t2.asVector();

            float resa = 0; // bei beiden Inv ist das Merkmal vorhanden
            float resb = 0; // nur bei Inv1 ist das Merkmal vorhanden
            float resc = 0; // nur bei Inv2 ist das Merkmal vorhanden
            float resd = 0; // bei keiner der Invarianten ist das Merkmal
                            // vorhanden
            float resm = 0;
            float res; // result

            for (int i = 0; i < vectorInv1.size(); i++) {
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) > 0)
                    resa += 1;
                if (vectorInv1.get(i) > 0 && vectorInv2.get(i) == 0)
                    resb += 1;
                if (vectorInv2.get(i) > 0 && vectorInv1.get(i) == 0)
                    resc += 1;
                if (vectorInv2.get(i) == 0 && vectorInv1.get(i) == 0)
                    resd += 1;
            }
            resm = resa + resb + resc + resd;
            res = (resa + resd) / resm;
            res = 1 - res;
            return res * 30;
        }
    }
}