/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.topology;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;

public class AvgClusteringCoefficients {

    /*
     * Wikipedia: The clustering coefficient C_i for a vertex v_i is given by
     * the proportion of links between the vertices within its neighborhood
     * divided by the number of links that could possibly exist between them.
     */
    int[][] adjacencyMatrixPlace, adjacencyMatrixTransition;
    HashMap<String, Integer> name2MatrixEntryPlace = new HashMap<String, Integer>();
    HashMap<String, Integer> name2MatrixEntryTransition = new HashMap<String, Integer>();
    HashMap<String, List<String>> graphPlace, graphTransition;
    PetriNet pn;

    public AvgClusteringCoefficients(PetriNet pn, HashMap<String, List<String>> graphPlace,  HashMap<String, List<String>> graphTransition) {
        this.pn = pn;
        this.graphPlace = graphPlace;
        this.graphTransition = graphTransition;
        this.adjacencyMatrixPlace = new int[pn.places().size()][pn.places().size()];
        this.adjacencyMatrixTransition = new int[pn.transitions().size()][pn.transitions().size()];
        fillMatrix();
    }

    public void fillMatrix() {
        Integer i = 0;
        for (Place n : pn.places()) {
            this.name2MatrixEntryPlace.put(n.<String> getProperty("name"), i);
            i++;
        }
        for (String from : this.graphPlace.keySet()) {
            for (String to : this.graphPlace.get(from)) {
                this.adjacencyMatrixPlace[name2MatrixEntryPlace.get(from)][name2MatrixEntryPlace.get(to)] = 1;
            }
        }
        
        i = 0;
        for (Transition n : pn.transitions()) {
            this.name2MatrixEntryTransition.put(n.<String> getProperty("name"), i);
            i++;
        }
        for (String from : this.graphTransition.keySet()) {
            for (String to : this.graphTransition.get(from)) {
                this.adjacencyMatrixTransition[name2MatrixEntryTransition.get(from)][name2MatrixEntryTransition.get(to)] = 1;
            }
        } 
    }

    public String compute() {
        HashMap<Integer, List<Double>> ccMapPlace = new HashMap<Integer, List<Double>>();
        HashMap<Integer, List<Double>> ccMapTransition = new HashMap<Integer, List<Double>>();
        for (String i : graphPlace.keySet()) {
            List<String> values = graphPlace.get(i);
            int n_i = values.size();
            if (n_i > 1) {
                Integer enumerator = 0;
                for (int j = 0; j < values.size(); j++) {
                    for (int k = j; k < values.size(); k++) {
                        Integer connected = adjacencyMatrixPlace[name2MatrixEntryPlace.get(values.get(j))][name2MatrixEntryPlace.get(values.get(k))]
                                           +adjacencyMatrixPlace[name2MatrixEntryPlace.get(values.get(k))][name2MatrixEntryPlace.get(values.get(j))];
                        enumerator = enumerator + connected;
                    }
                }
                Integer denominator = (n_i * (n_i - 1));
                double cc = (double) enumerator / (double) denominator;

                if (!((double) enumerator / (double) denominator > 1.0)) {
                    List<Double> vec = ccMapPlace.get(n_i);
                    if (vec == null) {
                        vec = new ArrayList<Double>();
                    }
                    vec.add(cc);
                    ccMapPlace.put(n_i, vec);
                } else {
                    throw new RuntimeException("Error in average clustering coefficient computation.");
                }
            }
            else {
                List<Double> vec = ccMapPlace.get(n_i);
                if (vec == null) {
                    vec = new ArrayList<Double>();
                }
                vec.add(0.0);
                ccMapPlace.put(n_i, vec);
            }
        }
        for (String i : graphTransition.keySet()) {
            List<String> values = graphTransition.get(i);
            int n_i = values.size();
            if (n_i > 1) {
                Integer enumerator = 0;
                for (int j = 0; j < values.size(); j++) {
                    for (int k = j; k < values.size(); k++) {
                        Integer connected = adjacencyMatrixTransition[name2MatrixEntryTransition.get(values.get(j))][name2MatrixEntryTransition.get(values.get(k))]
                                          +adjacencyMatrixTransition[name2MatrixEntryTransition.get(values.get(k))][name2MatrixEntryTransition.get(values.get(j))];
                        enumerator = enumerator + connected;
                    }
                }
                Integer denominator = (n_i * (n_i - 1));
                double cc = (double) enumerator / (double) denominator;

                if (!((double) enumerator / (double) denominator > 1.0)) {
                    List<Double> vec = ccMapTransition.get(n_i);
                    if (vec == null) {
                        vec = new ArrayList<Double>();
                    }
                    vec.add(cc);
                    ccMapTransition.put(n_i, vec);
                } else {
                    throw new RuntimeException("Error in average clustering coefficient computation.");
                }
            }
            else {
                List<Double> vec = ccMapTransition.get(n_i);
                if (vec == null) {
                    vec = new ArrayList<Double>();
                }
                vec.add(0.0);
                ccMapTransition.put(n_i, vec);
            }
        }        
        
        return print(ccMapPlace, ccMapTransition);
    }

    public String print(HashMap<Integer, List<Double>> ccMapPlace, HashMap<Integer, List<Double>> ccMapTransition) {
        String degree = "x_1<-c(";
        String avgCC = "y_1<-c(";
        String boxPlot = "data <- data.frame()\n"; 
            
        List<Integer> orderNeighbours = new ArrayList<Integer>(ccMapPlace.keySet());  
        Collections.sort(orderNeighbours);
        for (Integer i : orderNeighbours) {
            Double count = 0.0;
            for (Double d : ccMapPlace.get(i)) {
                boxPlot += "data <- rbind(data, c("+d+","+i+"))\n";
                count = count + d;
            }
            Double countAvg = count / (double) ccMapPlace.get(i).size();
            degree = degree.concat(String.valueOf(i) + ", ");
            avgCC = avgCC.concat(String.valueOf(countAvg) + ", ");
        }
        
        degree = degree.substring(0, degree.lastIndexOf(","));
        degree = degree.concat(")");
        avgCC = avgCC.substring(0, avgCC.lastIndexOf(","));
        avgCC = avgCC.concat(")");
        boxPlot += "colnames(data) <- c(\"cc\", \"neighbours\")\nboxplot(cc ~ neighbours, data, xlab = \"# of neighbors\", ylab = \"clustering coefficient (metabolites)\")";
        
        String output = degree + "\n" + avgCC 
                      + "\nbarplot(y_1, names.arg = x_1, xlab = \"# of neighbors\", ylab = \"average cluster coefficent (metabolites)\")"
                      + "\n"+boxPlot;;
        
        
        degree = "x_2<-c(";
        avgCC = "y_2<-c(";
        boxPlot = "data <- data.frame()\n"; 
        
        orderNeighbours = new ArrayList<Integer>(ccMapTransition.keySet());
        Collections.sort(orderNeighbours);
        for (Integer i : orderNeighbours) {
            Double count = 0.0;
            for (Double d : ccMapTransition.get(i)) {
                boxPlot += "data <- rbind(data, c("+d+","+i+"))\n";
                count = count + d;
            }
            Double countAvg = count / (double) ccMapTransition.get(i).size();
            degree = degree.concat(String.valueOf(i) + ", ");
            avgCC = avgCC.concat(String.valueOf(countAvg) + ", ");
        }
        
        degree = degree.substring(0, degree.lastIndexOf(","));
        degree = degree.concat(")");
        avgCC = avgCC.substring(0, avgCC.lastIndexOf(","));
        avgCC = avgCC.concat(")");
        boxPlot += "colnames(data) <- c(\"cc\", \"neighbours\")\nboxplot(cc ~ neighbours, data, xlab = \"# of neighbors\", ylab = \"clustering coefficient (reactions)\")";
        
        output += "\n" + degree + "\n" + avgCC 
                + "\nbarplot(y_2, names.arg = x_2, xlab = \"# of neighbors\", ylab = \"average cluster coefficent (reactions)\")"
                + "\n"+boxPlot;;        
        
        return output;
    }
}